#lang racket

(require racket/tcp)

(provide (all-defined-out))

(define PORT   4567)
(define TRIES  3)     ;; how many times should register try to connect to the server 
(define PAUSE  1/2)     ;; # secs to wait between attempts to connect to server 

(define (sexp? x)
  (cond
    [(empty? x) true]
    [(string? x) true]
    [(symbol? x) true]
    [(number? x) true]
    [(boolean? x) true]
    [(char? x) true]
    [(pair? x) (and (list? x) (andmap sexp? x))]
    [(and (struct? x) (prefab-struct-key x)) (for/and ((i (struct->vector x))) (sexp? i))]
    [else false]))

(define tcp-eof (gensym 'tcp-eof))

;; Any -> Boolean 
(define (tcp-eof? a) (eq? tcp-eof a))

;; OutPort Sexp -> Void
(define (tcp-send out msg)
  (write msg out)
  (newline out)
  (flush-output out))

;; InPort -> Sexp
(define (tcp-receive in)
  (with-handlers ((exn? (lambda (x) (raise tcp-eof))))
    (define x (read in))
    (if (eof-object? x) 
        (raise tcp-eof)
        (begin
          (read-line in) ;; read the newline 
          x))))

;; InPort OutPort (U #f String) -> Void 
;; register with the server, send the given name or make up a symbol 
(define (tcp-register in out name)
  (define msg `(REGISTER ((name ,(if name name (gensym 'world))))))
  (tcp-send out msg)
  (define ackn (tcp-receive in))
  (unless (equal? ackn '(OKAY))
    (raise tcp-eof)))

(define (register-with-host name register callback)
  (define *rec* (make-custodian))
  (define FMT "\nworking off-line\n")
  (define FMTtry (string-append "unable to register with ~a after ~s tries" FMT))
  (define FMTcom (string-append "unable to register with ~a due to protocol problems" FMT))
  ;; Input-Port -> [-> Void]
  ;; create closure (for thread) to receive messages and signal events
  (define (RECEIVE in)
    (define (RECEIVE)
      (sync 
       (handle-evt
        in
        (lambda (in) 
          (define dis "the universe disappeared")
          ((with-handlers ((tcp-eof? 
                            (lambda (e) (callback dis))))
             ;; --- "the universe disconnected" should come from here ---
             (define msg (tcp-receive in))
             (cond
               [(sexp? msg) (callback msg) RECEIVE] ;; break loop if EOF
               [else (error 'RECEIVE "sexp expected, received: ~e" msg)])))))))
    RECEIVE)
  ;; --- now register, obtain connection, and spawn a thread for receiving
  (parameterize ([current-custodian *rec*])
    ;; try to register with the server n times 
    (let try ([n TRIES])
      (printf "trying to register with ~a ...\n" register)
      (with-handlers ((tcp-eof? (lambda (x) (printf FMTcom register)))
                      (exn:fail:network? 
                       (lambda (x)
                         (if (= n 1) 
                             (printf FMTtry register TRIES)
                             (begin (sleep PAUSE) (try (- n 1)))))))
        (define-values (in out) (tcp-connect register PORT))
        (tcp-register in out name)
        (printf "... successful registered and ready to receive\n")
        (values out (thread (RECEIVE in)))))))
